#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <string.h>
#include <fcntl.h>

#include "clogger.h"
#include "libcr.h"
#include "cr_sigforward.h"

#define READMSG_TIMEOUT_SEC 1

#if defined(USING_CR_RESTART)
    char mpirun_command[] =  "cr_run";
    char restart_command[] = "cr_restart";
#else
    char mpirun_command[] =  "mpirun";
    char restart_command[] = "ompi-restart";
#endif

char *invalid_restart_identifier="Failed to initialize a restart identifier!";

enum msg_e {
    no_msg = 0,
    msg_start,
    msg_checkpoint,
    msg_continue,
    msg_restart,
    msg_crerror,
    msg_child_exited,
    msg_child_error,
};

static char *msg_names[] = {
    "No message",
    "msg_start",
    "msg_checkpoint",
    "msg_continue",
    "msg_restart",
    "msg_crerror",
    "msg_child_exited",
    "msg_child_error",
};

enum child_state_e {
    child_state_abort = 0,
    child_state_new,
    child_state_running,
    child_state_checkpointing,
    child_state_checkpoint_self,
    child_state_checkpoint_done,
    child_state_complete,
    child_state_restarting,
};

static char *state_names[] = {
    "child_state_abort",
    "child_state_new",
    "child_state_running",
    "child_state_checkpointing",
    "child_state_checkpoint_self",
    "child_state_checkpoint_done",
    "child_state_complete",
    "child_state_restarting",
};

struct channel_struct {
    sem_t put_sem;
    sem_t take_sem;
    enum msg_e the_msg;
};

#define MAX_RESTARTID_LENGTH 256

/*
 * We use these when parsing the output of the ompi-checkpoint command.
 *
 * NOTE:  Set the length in the scanf string to MAX_RESTARTID_LENGTH-1.  
 * We don't want to overrun the structure!
 */
#define SNAPSHOT_SCANF_FORMAT "Snapshot Ref.: %3d %255s\n"
#define SNAPSHOT_SCANF_COUNT  2

#define OUTPUT_BUFFER_SIZE 4096

/*
 * context_struct gets passed to each state function.  
 * Used to pass information between states.
 */
struct child_context_struct {
    int orterun_argc;
    char **orterun_argv;
    int restart_argc;
    char **restart_argv;
    char **envp;
    int *pipe;
    int child_exit_status;
    int child_pid;
    char restart_identifier[MAX_RESTARTID_LENGTH];
    pthread_t thread;
};

/*
 * waitpid struct is used to pass the return value(s) of waitpid() through
 * a pipe
 */
struct waitpid_struct {
    int wait_errno;
    pid_t pid;
    int status;
};


/*
 * global variables for communicating with the worker threads and signal
 * handlers
 */
struct channel_struct event_channel;

/*
 * the checkpoint acquires checkpoint_handler_lock before entry to cr_checkpoint()
 */
static pthread_mutex_t checkpoint_handler_lock = PTHREAD_MUTEX_INITIALIZER;

/*
 * the checkpoint acquires checkpoint_handler_lock before entry to cr_checkpoint()
 */
static pthread_mutex_t checkpoint_self_lock = PTHREAD_MUTEX_INITIALIZER;

/*
 * the thread that execs mpirun takes this lock before returning the
 * commands exit status.  This ensures that exits don't have to be dealt
 * with while we're in the checkpoint_self state.
 */
static pthread_mutex_t exit_lock = PTHREAD_MUTEX_INITIALIZER;

/*
 * we use this pipe to notify ourselves of restart and continue events
 */
static int child_event_pipe[2];

/*
 * The grandchild reads this pipe to determine when the greatgrandchild 
 * (mpirun or ompi-restart) has exited.  The write() is done in 
 * sigchild_handler(), in the same process.
 */
static int sigchild_pipe[2];

/*
 * signal mask
 */
static sigset_t my_sigmask;

/*
 * put and take
 *
 * these are simple send and receive operations on a channel.
 *
 * put is send, and blocks if the buffer is full.
 * take is a receive, and blocks if the buffer is empty.
 *
 * we use two semaphores to protect the (single-entry) message queue.
 */
static void put(struct channel_struct *channel, enum msg_e sent_msg)
{
    sem_wait(&channel->put_sem);
    channel->the_msg = sent_msg;
    sem_post(&channel->take_sem);
}

static enum msg_e take(struct channel_struct *channel)
{
    enum msg_e ret;

    sem_wait(&channel->take_sem);
    ret = channel->the_msg;
    sem_post(&channel->put_sem);

    return ret;
}

static void channel_init(struct channel_struct *channel)
{
    sem_init(&channel->take_sem, 1, 0);
    sem_init(&channel->put_sem, 1, 1);
}

static void debug_print_argv(int argc, char *const argv[])
{
    int i;
    size_t buflen = 4096;
    char buf[buflen];
    int off;
    char *s;

    s = buf;
    off = 0;

    for (i=0; i<argc; ++i) {
        off += snprintf(s+off, buflen-off, "%s", argv[i]);
        if (i<argc-1) { off += snprintf(s+off, buflen-off, " "); }
    }

    CLOG_DEBUG(buf);
}

static void debug_print_execvp_argv(char *const argv[])
{
    int i;
    for (i=0; argv[i] != NULL; ++i);
    debug_print_argv(i, argv);
}

/*
 * This is a thread context callback registered with BLCR.
 * 
 * It coordinates with the main thread through the put and checkpoint_handler_lock
 * functions.
 *
 * The basic scheme is:
 * 
 * notify the main thread that a checkpoint is beginning
 * wait on checkpoint_handler_lock until we're unblocked by the main thread
 * callback to BLCR to checkpoint ourselves
 * notify the main thread that the checkpoint is complete
 */
static int my_callback(void* arg)
{
    int * rc_ptr = (int *)arg;
    int rc;
    enum msg_e msg;

    put(&event_channel, msg_checkpoint);

    /* these two locks act as a barrier against the main thread.  They ensure
     * that this thread cannot run until the main thread unlocks
     * checkpoint_handler_lock, and the main thread blocks until we signal it 
     * with checkpoint_self_lock */
    pthread_mutex_lock(&checkpoint_self_lock);
    pthread_mutex_lock(&checkpoint_handler_lock);
    rc = cr_checkpoint(0);
    pthread_mutex_unlock(&checkpoint_self_lock);
    pthread_mutex_unlock(&checkpoint_handler_lock);

    if (rc_ptr != NULL) {
        *rc_ptr = rc;
    }

    if (rc > 0) {
        msg = msg_restart;
    } else if (rc == 0) {
        msg = msg_continue;
    } else {
        msg = msg_crerror;
    }

    /* wake up the main thread */
    put(&event_channel, msg);

    return 0;
}

int set_fd_cloexec(int fd)
{
    int retval;
    int flags;

    flags = fcntl(fd, F_GETFD);
    if (flags < 0) {
        CLOG_PERROR("fcntl");
        goto out;
    } 

    flags |= FD_CLOEXEC;
    retval =  fcntl(fd, F_SETFD, flags);
    if (retval < 0) {
        CLOG_PERROR("fcntl");
    } 

out:
    return retval;
}

int pipe_cloexec(int *filedes)
{
    int retval;

    retval = pipe(filedes);
    if (retval < 0) {
        CLOG_PERROR("pipe");
        goto out;
    }

    retval = set_fd_cloexec(filedes[0]);
    if (retval < 0) {
        goto out_close;
    }

    retval = set_fd_cloexec(filedes[1]);
    if (retval < 0) {
        CLOG_PERROR("fcntl");
        goto out_close;
    }

out:
    return retval;

out_close:
    close(filedes[0]);
    close(filedes[1]);
    return retval;
}

/*
 * Here's another system()-like function.  This thing is basically backticks
 * like in the shell.  It spawns a new process, and saves its standard
 * output to a buffer.  If the output of the child process is too large,
 * we deadlock against it. -- see the XXX below.  :(
 *
 * (This can't be right...  There has to be a better way...  Can't we read
 * until the child closes the thing and wait for an EOF?)
 */
static int
get_child_output(const char *path, char *const argv[], char *output, 
    int output_length, int *cmd_exit_status)
{
    int retval;
    int filedes[2];
    int pid;
    int wpid;

    debug_print_execvp_argv(argv);

    /* open pipes to child */
    retval = pipe_cloexec(filedes);
    if (retval < 0) {
        goto out;
    }

    pid = fork();
    if (pid > 0) {
        /* parent */

        /* close the write end of the pipe. */
        close(filedes[1]);

        /* read standard output */
        retval = read(filedes[0], output, output_length);
        if (retval < 0) {
            CLOG_PERROR("read");
        }

        /* XXX:  I'm not quite sure this is right.  If the child fails to
         * exit, since it wrote too much data to the pipe, then we'll wait
         * forever here.
         *
         * The alternative is to do the read() first, but then if the child
         * exits without having written anything, we wait forever also.
         */
        wpid = waitpid(pid, cmd_exit_status, 0);
        if (wpid < 0) {
            CLOG_PERROR("waitpid");
        }

        /* check the return code.  If nonzero we'll assume that something
         * went wrong and abort. otherwise we'll continue and read the
         * things standard output. */
        if (*cmd_exit_status != 0) {
            CLOG_VERBOSE("cmd_exit_status = %d\n", *cmd_exit_status);
            goto out_close;
        }

        CLOG_VERBOSE("Read %d bytes from child process:", retval);
        CLOG_VERBOSE("%s", output);

    } else if (pid == 0) {
        /* child */

        /* might as well allow this thing to receive signals */
        reset_signals(&my_sigmask);

        /* close the read end of the pipe. */
        close(filedes[0]);

        /* connect standard output to the write pipe */
        dup2(filedes[1], 1);

        /* run the command */
        execvp(path, argv);

        /* only reached on error, exit and return to caller. */
        CLOG_PERROR("execvp");
        exit(-1);
    } else {
        CLOG_PERROR("fork");
        retval = pid;
        goto out_close;
    }

out_close:
    close(filedes[0]);
    close(filedes[1]);
out:
    return retval;
}

static int wait_for_read(int fd1, int fd2)
{
    fd_set rfds;
    int retval;
    int maxfd;

    maxfd = (fd1 > fd2) ? fd1+1 : fd2+1;

    /* This is written as a loop, since later, we'll want to loop
     * until either a restart occurs, or the fd becomes readable.
     */
    do {
        /* Watch fd to see when it's readable.  We initialize these on
         * every pass, since the values of the fd_set arrays are undefined
         * after an error. */
        FD_ZERO(&rfds);
        FD_SET(fd1, &rfds);
        FD_SET(fd2, &rfds);

        CLOG_VERBOSE("Entering select(%d, [%d, %d])", maxfd, fd1, fd2);
        retval = select(maxfd, &rfds, NULL, NULL, NULL);
        CLOG_VERBOSE("select returned %d", retval);

        if (retval == -1) {
            if (errno == EINTR) {
                /* probably just a signal, ignore it until we're told to
                 * do something by the main thread. */
                CLOG_VERBOSE("ignoring EINTR during select");
                ;
            } else {
                CLOG_PERROR("select");
                goto out;
            }
        } else if (retval) {
            if (FD_ISSET(fd1, &rfds)) {
                retval = fd1;
                goto out;
            } else if (FD_ISSET(fd2, &rfds)) {
                retval = fd2;
                goto out;
            } else {
                CLOG_VERBOSE("select returned unexpectedly");
                retval = -1;
            }
        } else {
            CLOG_VERBOSE("unexpected timeout from select");
            /* This should only happen on a timeout */
            goto out;
        }
    } while (1);

out:
    return retval;
}

static int readmsg(int fd, enum msg_e *msg)
{
    ssize_t bytes;
    fd_set rfds;
    int maxfd = fd+1;
    struct timeval timeout;
    int retval;

    do {
        FD_ZERO(&rfds);
        FD_SET(fd, &rfds);

        timeout.tv_sec = READMSG_TIMEOUT_SEC;
        timeout.tv_usec = 0;

        CLOG_VERBOSE("Entering select(%d, [%d])", maxfd, fd);
        retval = select(maxfd, &rfds, NULL, NULL, &timeout);
        CLOG_VERBOSE("select returned %d", retval);

        if (retval == 1) {
            /* return the message to the caller */
            bytes = read(fd, msg, sizeof(*msg));
            if (bytes != sizeof(*msg)) {
                retval = -1;
            } else {
                retval = 0;
            }
            goto out;
        } else if (retval == 0) {
            /* timed out, return an error */
            retval = -1;
            errno = ETIMEDOUT;
        } else { /* retval < 0 */
            if (errno == EINTR) {
                continue;
            }
        }
        goto out;
    } while(1);

out:
    return retval;
} 
 
static int read_wait_status(int child_fd, int event_fd,
                            struct waitpid_struct *wait_status)
{
    ssize_t bytes;
    enum msg_e event_msg;
    enum msg_e child_msg = msg_child_error;
    int retval;

    bytes = read(child_fd, wait_status, sizeof(*wait_status));
    CLOG_VERBOSE("read %d bytes from grandchild process on fd %d",
                 bytes, child_fd);
    if (bytes == 0) {
        /* bytes == 0 (restart)
         *
         * if we restart, child_fd is now attached to /dev/null, because
         * there is now no process attached to it.  So the read returns 0.
         *
         * We close the descriptor, then expect to be notified of the restart
         * on event_fd
         */
        close(child_fd);

        /* now, get a message on the event_fd descriptor, there
         * had better be a restart message there.  If NOT, then 
         * something is seriously wrong, and we try to abort nicely.
         */
        retval = readmsg(event_fd, &event_msg);
        if (retval < 0) {
            if (errno == ETIMEDOUT) {
                fprintf(stderr, "Timed out, expecting restart message.\n");
            }

            /* Don't forget to wake up the main thread! */
            goto out_exited;
        }

        if (event_msg == msg_restart) {
            retval = -1;
            errno = EINTR;
        } else {
            CLOG_VERBOSE
                ("Received message %d when expecting %d, aborting.",
                 event_msg, msg_restart);
            goto out_exited;
        }

        goto out;
    } else if (bytes != sizeof(*wait_status)) {
        fprintf(stderr, "read unexpectedly returned %ld\n", (long) bytes);
        wait_status->pid = -1;
        goto out_exited;
    }

    CLOG_VERBOSE("wait_status.pid = %d", wait_status->pid);
    CLOG_VERBOSE("wait_status.status = %d", wait_status->status);
    CLOG_VERBOSE("wait_status.wait_errno = %d", wait_status->wait_errno);

    retval = wait_status->pid;
    child_msg = msg_child_exited;

out_exited:
    pthread_mutex_lock(&exit_lock);
    put(&event_channel, child_msg);
    pthread_mutex_unlock(&exit_lock);
out:
    return retval;
}

static int get_handler_status(int child_fd, int event_fd)
{
    ssize_t bytes;
    enum msg_e msg;
    int retval = -1;

    CLOG_VERBOSE("checkpoint handler says something...");
    bytes = read(event_fd, &msg, sizeof(msg));
    CLOG_VERBOSE("read %d bytes from fd %d", bytes, event_fd);
    if (bytes < 0) {
        /* this is bad */
        CLOG_PERROR("read");
        goto out;
    } else if (bytes != sizeof(msg)) {
        fprintf(stderr, "short read on event_fd!\n");
        goto out;
    }

    if (msg == msg_continue) {
        CLOG_VERBOSE("continuing from checkpoint");
        /* wait for next event */
        retval = 0;
    } else if (msg == msg_restart) {
        CLOG_VERBOSE("restarting from checkpoint");

        /* the process on the other end exited, so close this. */
        close(child_fd);
        CLOG_VERBOSE("closed fd %d", child_fd);

        /* XXX: Probably unnecessary */
        errno = EINTR;
    } else if (msg == msg_crerror) {
        /* fatal */
        CLOG_VERBOSE("error from checkpoint");
    } else {
        CLOG_VERBOSE("unexpected msg from checkpoint");
    }

out:
    return retval;
}

/* wait for child process to exit */
static int wait_for_child(int child_pid)
{
    int retpid;
    int child_status;

    CLOG_VERBOSE("waiting for child process %d to exit", child_pid);
    retpid = waitpid(child_pid, &child_status, 0);
    if (retpid != child_pid) {
        fprintf(stderr, "waitpid returned %d unexpectedly!\n", retpid);
        return retpid;
    }

    return child_status;
}

/* This is roughly what we're trying to do
 *
 * wait_status.pid = 
 *     waitpid(pid_of_great_grandchild, &wait_status.status, 0); 
 *
 * except that the wait is performed by another process, who
 * passes the result back to us through a pipe, and also, we can be 
 * checkpointed while waiting, in which case the process writing the pipe
 * will suddenly disappear!
 */
static void wait_command(int child_fd, int event_fd,
                         struct waitpid_struct *wait_status)
{
    /* parent */
    int done = 0;
    int pid;
    int retval;

    do {
        retval = wait_for_read(event_fd, child_fd);
        if (retval == child_fd) {
            pid = read_wait_status(child_fd, event_fd, wait_status);
            done = 1;
        } else if (retval == event_fd) {
            done = get_handler_status(child_fd, event_fd);
        } else {
            /* error from select? */
            CLOG_PERROR("wait_for_read");
        }
    } while (!done);
}

/*
 * wait_for_greatgrandchild
 *
 * - writes greatgrandchild_pid to parent_fd,
 * - calls waitpid() on greatgrandchild_pid, and 
 * - writes the resulting wait_status back to parent_fd
 */
static void wait_for_greatgrandchild(int parent_read_fd, int parent_write_fd,
                                     int greatgrandchild_pid)
{
    /* grandchild */
    ssize_t bytes;
    struct waitpid_struct wait_status;
    int nextfd;

    /* inform the parent of the greatgrandchild's PID */
    bytes =
        write(parent_write_fd, &greatgrandchild_pid,
              sizeof(greatgrandchild_pid));
    if (bytes != sizeof(greatgrandchild_pid)) {
        CLOG_PERROR("write");
    }

    /* forward signals sent to us to the greatgrandchild process -- except
     * SIGCHLD, which we want. */
    forward_signals(&my_sigmask, greatgrandchild_pid);

    /* wait for the great-grandchild to exit (or for our parent to exit) */
    nextfd = wait_for_read(parent_read_fd, sigchild_pipe[0]);
    if (nextfd == sigchild_pipe[0]) {
        /* the child has exited, no reason to forward signals anymore */
        reset_signals(&my_sigmask);

        bytes = read(sigchild_pipe[0], &wait_status, sizeof(wait_status));
        if (bytes != sizeof(wait_status)) {
            CLOG_PERROR("read");
        }

        /* write the ggchild's exit_status to the pipe */
        bytes = write(parent_write_fd, &wait_status, sizeof(wait_status));
        if (bytes != sizeof(wait_status)) {
            CLOG_PERROR("write");
        }
    } else if (nextfd == parent_read_fd) {
        /* detected abnormal termination of parent */
        fprintf(stderr, "Abnormal termination of parent!\n");

        /* send a HUP to the greatgrandchild */
        kill(greatgrandchild_pid, SIGHUP);

        /* we could try to wait for the child, but someone may be trying
         * to kill us, so we'll exit.  We have no one to return the error code
         * to, so I think this is the best we can do */
    }
}

static void wait_for_parent_exit(void) 
{
    do {
        sched_yield();
    } while (getppid() != 1);
}


void sigchild_handler(int signum)
{
    struct waitpid_struct wait_status;
    ssize_t bytes;

    wait_status.pid = waitpid(-1, &wait_status.status, 0);
    wait_status.wait_errno = errno;

    /* send back the result */
    bytes = write(sigchild_pipe[1], &wait_status, sizeof(wait_status));
    if (bytes < 0) {
        /* not sure how (or whether) to try and handle this.  
         * It means that we couldn't notify ourselves of the
         * child exiting.
         *
         * So we're going to close this.  The select() later should
         * return EINTR and be retried, and wait_for_greatgrandchild() on
         * the next iteration should find the EBADF and abort.
         */
        close(sigchild_pipe[1]);
    }
}

static void exec_grandchild(int parent_read_fd, int parent_write_fd,
                            const char *filename, char *const argv[])
{
    /* grandchild */
    int greatgrandchild_pid;
    int retval;
    struct sigaction sa;

    /* Wait for child (our parent) to exit, so init is our parent process */
    wait_for_parent_exit();

    /* XXX: Call setpgid() optionally */

    /* XXX: Call setsid() optionally */

    /* create a pipe to use for SIGCHLD notification */
    retval = pipe_cloexec(sigchild_pipe);
    if (retval < 0) {
        /* if we prematurely close, the parent sees the error */
        exit(-1);
    }

    /* install SIGCHLD handler */
    sa.sa_handler = sigchild_handler;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGCHLD, &sa, NULL);

    greatgrandchild_pid = fork();
    if (greatgrandchild_pid > 0) {
        /* grandchild */
        wait_for_greatgrandchild(parent_read_fd, parent_write_fd,
                                 greatgrandchild_pid);

        exit(-1);
        /* not reached */
    } else if (greatgrandchild_pid == 0) {
        /* greatgrandchild */

        /* probably unnecessary */
        sa.sa_handler = SIG_DFL;
        sigaction(SIGCHLD, &sa, NULL);

        /* close pipes back to the parent */
        close(parent_write_fd);
        close(parent_read_fd);

        /* don't need this either */
        close(sigchild_pipe[0]);
        close(sigchild_pipe[1]);

        /* install the default set of signal handlers and original sigset */
        reset_signals(&my_sigmask);

        /* finally, execvp the process we wanted */
        execvp(filename, argv);

        /* not reached */
        CLOG_PERROR("execvp");
    } else {
        /* unable to fork() */
        CLOG_PERROR("fork(grandchild)");

        /* 
         * original parent process is expecting us to write(), 
         * we can close prematurely to indicate an error.
         */
        close(parent_write_fd);

        /* Unnecessary, but why not? */
        close(parent_read_fd);

        exit(-1);
    }

    /* does not return */
}

static void exec_command(int parent_read_fd, int parent_write_fd,
                         const char *filename, char *const argv[])
{
    int grandchild_pid;

    grandchild_pid = fork();
    if (grandchild_pid > 0) {
        /* still in child: exit to remove grandchild from process tree */
        exit(0);
    } else if (grandchild_pid == 0) {
        exec_grandchild(parent_read_fd, parent_write_fd, filename, argv);
        /* not reached */
    } else {
        /* in child, but unable to create a grandchild
         * return an error to the parent */
        CLOG_PERROR("fork(child)");
        exit(-1);
    }

    /* does not return */
}

/*
 * like system() but not.  We'll accept SIGINT and the like, just to make
 * things more complicated, and we'll also try to return more (or just print)
 * more information when things go wrong
 */
static int isystem(const char *filename, char *const argv[],
                   char *const envp[], int *p_pid, int *exit_status,
                   struct channel_struct *notification_channel, 
                   int eventpipe[2])
{
    int pid;
    int retval;
    int filedes[2];
    int watchpipe[2];
    int child_status;
    struct waitpid_struct wait_status;
    ssize_t bytes;
    int gg_pid;

    /* open a pipe for communication with the GRAND-child */
    retval = pipe_cloexec(filedes);
    if (retval < 0) {
        return retval;
    }

    /* This is the reverse of the other pipe.  The grandchild interprets a
     * close of this pipe to mean abnormal termination of the parent.  So
     * we never write or read anything from it.  We just check it in the
     * grandchild for EOF, and close it in the parent after it looks like the
     * grandchild has exited.
     */
    retval = pipe_cloexec(watchpipe);
    if (retval < 0) {
        return retval;
    }

    pid = fork();
    if (pid > 0) {
        /* parent */

        /* close the write end of the pipe */
        close(filedes[1]);
        /* close the read of the the watchpipe */
        close(watchpipe[0]);

        /* wait for the child process to exit (unnecessary?) */
        child_status = wait_for_child(pid);
        if (child_status != 0) {
            fprintf(stderr, "wait_for_child unexpectedly returned %d!\n",
                    child_status);
        }

        /* read the greatgrandchild's PID */
        bytes = read(filedes[0], &gg_pid, sizeof(gg_pid));
        if (bytes != sizeof(gg_pid)) {
            fprintf(stderr,
                    "read unexpectedly returned %ld while waiting for child process\n",
                    (long) bytes);

            /* XXX: How do I get out of this? */

            /* XXX:  Need to put() to unblock the main thread! */
            /* XXX:  And need to set some kind of error flag... */
            /* XXX:  Try setting the gg_pid to an error code, and have the
             * main threads check after a start event */
        }
        CLOG_VERBOSE("read %d bytes from fd %d", bytes, filedes[0]);

        CLOG_VERBOSE("greatgrandchild PID is %d", gg_pid);
        if (p_pid != NULL) {
            *p_pid = gg_pid;
        }

        if (notification_channel != NULL) {
            put(notification_channel, msg_start);
        }

        wait_command(filedes[0], eventpipe[0], &wait_status);

        close(watchpipe[1]);

        *exit_status = wait_status.status;
        errno = wait_status.wait_errno;

    } else if (pid == 0) {
        /* child */

        /* close the read end of the pipe */
        close(filedes[0]);
        /* close the write end of the watchpipe */
        close(watchpipe[1]);

        /* also, close the pipe to the main thread */
        close(eventpipe[0]);
        close(eventpipe[1]);

        exec_command(watchpipe[0], filedes[1], filename, argv);
        /* not reached */
    } else {
        CLOG_PERROR("fork(parent)");
        return pid;
    }

    return retval;
}

/*
 * runmpi() - this is the thread that exec()'s mpirun, and waits for the
 * application to exit.  called back by pthread_create()
 */
static void *runmpi(void *buf)
{
    struct child_context_struct *my_args = (struct child_context_struct *) buf;
    char **orterun_argv = my_args->orterun_argv;
    int ret;

    debug_print_argv(my_args->orterun_argc, my_args->orterun_argv);

    /* spawn mpirun */
    ret = isystem(mpirun_command, orterun_argv, my_args->envp, 
            &my_args->child_pid, &my_args->child_exit_status, &event_channel,
            my_args->pipe);

    pthread_exit(&my_args->child_exit_status);
}

static void *restartmpi(void *buf)
{
    struct child_context_struct *my_args = (struct child_context_struct *) buf;
    int ret;
    char **restart_argv = my_args->restart_argv;
    int restart_argc = my_args->restart_argc;

    CLOG_VERBOSE("restarting child process.");

    my_args->restart_argv[restart_argc-1] = my_args->restart_identifier;

    debug_print_argv(restart_argc, restart_argv);

    ret = isystem(restart_command, restart_argv, my_args->envp, 
            &my_args->child_pid, &my_args->child_exit_status, &event_channel,
            my_args->pipe);

    pthread_exit(&my_args->child_exit_status);

}


/*
 * this is our panic state.  If a state transition fails, or we somehow
 * receive a forbidden event, we go here.
 */
static enum child_state_e 
cr_mpirun_abort(struct child_context_struct *child_context)
{
    CLOG_VERBOSE("cr_mpirun_abort");
    return child_state_abort;
}

/* child_new
 *
 * this just advances us into the running state
 *
 * the state machine is initialized into this state.  I suppose we could do
 * our initialization work in here, but right now this function is purely
 */
static enum child_state_e 
child_new(struct child_context_struct *child_context)
{
    enum child_state_e new_state = child_state_abort;
    enum msg_e msg;

    CLOG_VERBOSE("child_new");

    /* start the runmpi thread */
    pthread_create(&(child_context->thread), NULL, runmpi,
                   (void *) child_context);

    msg = take(&event_channel);
    CLOG_VERBOSE("received message %s", msg_names[msg]);
    if (msg == msg_start) {
        new_state = child_state_running;
        /* XXX: should we send these directly to the greatgrandchild? */
        forward_signals(&my_sigmask, child_context->child_pid);
    }

    return new_state;
}

static enum child_state_e 
child_restarting(struct child_context_struct *child_context)
{
    enum child_state_e new_state = child_state_abort;
    int *old_status;
    enum msg_e msg;

    CLOG_VERBOSE("child_restarting");

    CLOG_VERBOSE("waiting for old thread to exit");
    pthread_join(child_context->thread, (void *) &old_status);

    /* start the restartmpi thread */
    pthread_create(&(child_context->thread), NULL, restartmpi,
                   (void *) child_context);

    msg = take(&event_channel);
    CLOG_VERBOSE("received message %s", msg_names[msg]);
    if (msg == msg_start) {
        forward_signals(&my_sigmask, child_context->child_pid);
        new_state = child_state_running;
    }

    return new_state;
}


/* child_running
 *
 * this means that the job is running normally, and we're waiting.   
 */
static enum child_state_e 
child_running(struct child_context_struct *child_context)
{
    enum child_state_e new_state = child_state_abort;
    enum msg_e msg;

    CLOG_VERBOSE("child_running with PID %d", child_context->child_pid);

    msg = take(&event_channel);
    CLOG_VERBOSE("received message %s", msg_names[msg]);

    if (msg == msg_child_exited) {
        new_state = child_state_complete;
    } else if (msg == msg_checkpoint) {
        new_state = child_state_checkpointing;
    } else if (msg == msg_child_error) {
        /* something happened, try to kill the grandchild process */
        kill(child_context->child_pid, SIGHUP);
        CLOG_VERBOSE("kill(%d, SIGHUP)", child_context->child_pid);
    }

    return new_state;
}

/*
 * child_checkpointing
 *
 * on entry, we've received a request to checkpoint the child
 * on return from this routine the child will have checkpointed
 * our output state should be child_checkpoint_self
 *
 * if we expect the child to exit on the checkpoint (because we're killing
 * it) then we call take() and wait for the child_exit message from the
 * wait() thread before proceeding to child_checkpoint_self.
 *
 * note that we hold the exit lock on entry into child_checkpoint_self
 * to prevent that state from having to deal with process termination.
 */
static enum child_state_e 
child_checkpointing(struct child_context_struct *child_context)
{
#if defined(USING_CR_RESTART)
#define OMPI_CHECKPOINT_COMMAND "cr_checkpoint"
#else
#define OMPI_CHECKPOINT_COMMAND "ompi-checkpoint"
#endif
#define OMPI_CHECKPOINT_ARGC 2
#define ARGV_LENGTH 255
    enum child_state_e new_state = child_state_abort;
    int ompi_checkpoint_exitstatus;
    int retval;
    int checkpoint_argc = OMPI_CHECKPOINT_ARGC;
    char *checkpoint_argv[checkpoint_argc+1];
    char ompi_checkpoint_command[] = OMPI_CHECKPOINT_COMMAND;
    char child_pidstr[ARGV_LENGTH];
    int snapshot_number;

    int killing_the_child = 0;
    char *checkpoint_stdout;

    checkpoint_stdout = malloc(OUTPUT_BUFFER_SIZE * sizeof(char));
    if (checkpoint_stdout == NULL) {
        CLOG_PERROR("malloc");
        goto out;
    }

    CLOG_VERBOSE("child_checkpointing");

    /* XXX: Rewrite this into a call that uses popen() or dup()/read() to
     * after forking the child and reading out the checkpoint location
     */

    /* build an argv for the child */
    snprintf(child_pidstr, ARGV_LENGTH, "%d", child_context->child_pid);
    checkpoint_argv[0] = ompi_checkpoint_command;
    checkpoint_argv[1] = child_pidstr;
    checkpoint_argv[2] = (char *) NULL;

    retval = get_child_output(ompi_checkpoint_command, 
                              checkpoint_argv, 
                              checkpoint_stdout, 
                              OUTPUT_BUFFER_SIZE, 
                              &ompi_checkpoint_exitstatus);

    /* if unsuccessful, panic */
    if (retval < 0) {
        CLOG_FATAL("Bad return code %d during %s, aborting...", 
                  retval, ompi_checkpoint_command);
        goto out;
    } else if (ompi_checkpoint_exitstatus != 0) {
        CLOG_FATAL("Nonzero exit status %d from %s, aborting...", 
            ompi_checkpoint_exitstatus, ompi_checkpoint_command);
        goto out;
    }


#if defined(USING_CR_RESTART)
    /* XXX:  This scanf thing should be a subroutine call.  Would like to
     * isolate the ompi-checkpoint specific code.
     * 
     * Should have something in here like set_restart_id()
     */

    /* set the restart identifier */
    snprintf(child_context->restart_identifier, MAX_RESTARTID_LENGTH, 
             "context.%s", child_pidstr);
#endif

    /* Scan the ompi-checkpoint output for the snapshot handle */
    retval = sscanf(checkpoint_stdout, 
                    SNAPSHOT_SCANF_FORMAT,
                    &snapshot_number,
                    child_context->restart_identifier);
    if (retval != SNAPSHOT_SCANF_COUNT) {
        CLOG_FATAL("Unable to parse output of %s.  scanf returned %d but expecting %d.", 
                  ompi_checkpoint_command, retval, SNAPSHOT_SCANF_COUNT);
        goto out;
    }

    /* XXX: I'm too lazy to trim the newline... */
    CLOG_DEBUG("Snapshot Ref.: %3d %s", snapshot_number, child_context->restart_identifier);

    if (killing_the_child) {
        enum msg_e msg = take(&event_channel);

        CLOG_VERBOSE("Waiting for child to exit.");
        msg = take(&event_channel);

        /* make sure what happened is what we thought was going to happen */
        if (msg == msg_child_exited) {
            new_state = child_state_checkpoint_self;
        } else {
            CLOG_FATAL("Unexpectedly received message %d while waiting for child to terminate", msg_names[msg]);
            new_state = child_state_abort;
        }
    } else {
        /* ok, we'll let the child run, but if it DOES exit while we're
         * checkpointing ourselves we don't want to deal with that while our
         * own checkpoint handler is running, since dealing with both the
         * continue, restart, and error cases is just flat out weird in that
         * case. So we hold the exit lock to prevent the message from the
         * waiting thread to be noticeable until we enter the done state.
         */

         pthread_mutex_lock(&exit_lock);
    }
    
    new_state = child_state_checkpoint_self;

out:
    return new_state;
}

static enum child_state_e 
child_checkpoint_self(struct child_context_struct *child_context)
{
    enum child_state_e new_state = child_state_abort;

    /* unblock our checkpoint handler so we can checkpoint*/
    pthread_mutex_unlock(&checkpoint_handler_lock);

    /* wait for the checkpoint thread to wake us */
    pthread_mutex_lock(&checkpoint_self_lock);

    /* now set up for the next checkpoint.  we acquire the handler lock to
     * prevent the handler from continuing until the next checkpoint, and
     * release our own lock so we can wait for the handler again. */
    pthread_mutex_lock(&checkpoint_handler_lock);
    pthread_mutex_unlock(&checkpoint_self_lock);

    /* proceed to checkpoint done */
    new_state = child_state_checkpoint_done;

    return new_state;
}

static enum child_state_e 
child_checkpoint_done(struct child_context_struct *child_context)
{
    enum child_state_e new_state = child_state_abort;
    enum msg_e msg;
    ssize_t bytes;

    CLOG_VERBOSE("child_checkpoint_done");

    /* unlock the checkpoint handler so we can process an exit event if
     * one occurred */
    pthread_mutex_unlock(&exit_lock);

    msg = take(&event_channel);
    switch (msg) {
    case msg_continue:
        /* nothing to do */
        new_state = child_state_running;
        break;
    case msg_restart:
        /* the runmpi thread will restore us to a running state. */
        new_state = child_state_restarting;
        break;
    case msg_crerror:
        new_state = child_state_abort;
        break;
    case msg_child_exited:
        /* the child exited while we were checkpointing */
        new_state = child_state_complete;
        break;
    default:
        new_state = child_state_abort;
        break;
    }

    /* notify the child thread */
    CLOG_VERBOSE("notifying child thread");
    bytes = write(child_event_pipe[1], &msg, sizeof(msg));
    if (bytes != sizeof(msg)) {
        CLOG_VERBOSE("failed to notify child thread");
    } else if (bytes < 0) {
        CLOG_PERROR("write");
    }

    return new_state;
}

static void usage(void)
{
    char *basename = "cr_mpirun";

    fprintf(stderr,
"Usage: %s [OPTION]...  [--] [PROGRAM]...\n"
"Start the given program using Open RTE and accept BLCR checkpoint requests\n"
"\n"
"   -am <arg0>            Aggregate MCA parameter set file list\n"
"   --app                 Provide an appfile to mpirun\n"
"   --fork                Fork off a new process which is the restarted\n"
"                         process instead of replacing orte_restart\n"
"   -gmca|--gmca <arg0> <arg1>\n"
"                         Pass global MCA parameters that are applicable to\n"
"                         all contexts (arg0 is the parameter name; arg1 is\n"
"                         the parameter value)\n"
"   -h|--help             Print this message\n"
"   -hostfile|--hostfile <arg0>\n"
"                         Provide a hostfile\n"
"   -machinefile|--machinefile <arg0>\n"
"                         Provide a hostfile\n"
"   -p|--preload          Preload the checkpoint files before restarting\n"
"                         (Default = Disabled)\n"
"   -s|--seq <arg0>       The sequence number of the checkpoint to start\n"
"                         from. (Default: -1, or most recent)\n"
"   -v|--verbose          Be verbose.  may be specified more than once\n"
,
    basename);
}

/*
 * parse_args
 * 
 * constructs the orterun_argv and restart_argv arrays
 *
 * returns -1 on error, 1 if the user asked for help, and 0 otherwise
 */
static int parse_args(int argc, char *argv[], int *orterun_argp, 
    char *orterun_argv[], int *restart_argp, char *restart_argv[])
{
    int i;
    int optend;
    int restart_argc;
    int orterun_argc;
    int rc = -1;

    restart_argc = 1;
    orterun_argc = 1;
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
            /* --help */
            rc = 1; 
            goto out;
        } else if (!strcmp(argv[i], "-hostfile")
                   || !strcmp(argv[i], "--hostfile")
                   || !strcmp(argv[i], "-machinefile")
                   || !strcmp(argv[i], "--machinefile")) {
            /* --hostfile: pass to both mpirun and ompi-restart */
            if (i+1 >= argc) {
                goto out;
            }

            restart_argv[restart_argc++] = argv[i];
            restart_argv[restart_argc++] = argv[i+1];
            orterun_argv[orterun_argc++] = argv[i];
            orterun_argv[orterun_argc++] = argv[i+1];

            i += 1;
        } else if (!strcmp(argv[i], "-s") || !strcmp(argv[i], "--seq")) {
            /* --seq: pass to ompi-restart */

            restart_argv[restart_argc++] = argv[i];
            restart_argv[restart_argc++] = argv[i+1];

            if (i+1 >= argc) {
                goto out;
            }
            i += 1;
        } else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--verbose")) {
            /* --verbose: pass to both mpirun and ompi-restart */
            restart_argv[restart_argc++] = argv[i];
            orterun_argv[orterun_argc++] = argv[i];
            if (clog_get_trace_level() == CLOG_LEVEL_DEBUG) {
                clog_set_trace_level(CLOG_LEVEL_VERBOSE);
            } else {
                clog_set_trace_level(CLOG_LEVEL_DEBUG);
            }
        } else if (!strcmp(argv[i], "-p") || !strcmp(argv[i], "--preload")) {
            /* --preload: pass to ompi-restart */
            restart_argv[restart_argc++] = argv[i];
        } else if (!strcmp(argv[i], "-am")) {
            /* --am: pass to both mpirun and ompi-restart */
            if (i+1 >= argc) {
                goto out;
            }

            restart_argv[restart_argc++] = argv[i];
            restart_argv[restart_argc++] = argv[i+1];
            orterun_argv[orterun_argc++] = argv[i];
            orterun_argv[orterun_argc++] = argv[i+1];

            i += 1;
        } else if (!strcmp(argv[i], "-gmca") || !strcmp(argv[i], "--gmca")) {
            /* --gmca: pass to mpirun and ompi-restart */
            if (i+2 >= argc) {
                goto out;
            }

            restart_argv[restart_argc++] = argv[i];
            restart_argv[restart_argc++] = argv[i+1];
            restart_argv[restart_argc++] = argv[i+2];
            orterun_argv[orterun_argc++] = argv[i];
            orterun_argv[orterun_argc++] = argv[i+1];
            orterun_argv[orterun_argc++] = argv[i+2];

            i += 2;
        } else if (!strcmp(argv[i], "--app")) {
            /* --app: pass to mpirun */
            if (i+1 >= argc) {
                goto out;
            }

            orterun_argv[orterun_argc++] = argv[i];
            orterun_argv[orterun_argc++] = argv[i+1];

            i += 1;
        } else if (!strcmp(argv[i], "--fork")) {
            /* --fork: pass to ompi-restart */
            restart_argv[restart_argc++] = argv[i];
        } else if (!strcmp(argv[i], "--")) {
            /* program arguments -- stop processing.. */
            break;
        } else if (argv[i][0] == '-') {
            /* hmm...  this doesn't look like one of our arguments */
            CLOG_WARN("Unknown switch: '%s', "
                      "remaining arguments will be passed only to mpirun",
                      argv[i]);
            CLOG_VERBOSE("Bad argument encountered at position %d", i);
            break;
        } else {
            /* We're done, pass everything else to the program. */
            break;
        }
    }

    orterun_argv[orterun_argc] = (char *) NULL;
    restart_argv[restart_argc++] = invalid_restart_identifier;
    restart_argv[restart_argc] = (char *) NULL;

    /* take any remaining arguments as the program name, followed by program
     * arguments, and pass these to mpirun */
    optend = i;
    for (i = optend; i < argc; i++) {
        orterun_argv[orterun_argc++] = argv[i];
    }

    restart_argv[0] = restart_command;
    orterun_argv[0] = mpirun_command;

    *restart_argp = restart_argc;
    *orterun_argp = orterun_argc;

    rc = 0;

out:
    return rc;
}

int main(int argc, char *argv[], char *envp[])
{
    int rc;
    enum child_state_e state;
    struct child_context_struct child_context;
    int *old_status = &rc;

    /* set up the state table */
    enum child_state_e (*handle[])(struct child_context_struct *) = {
        cr_mpirun_abort,
        child_new,
        child_running,
        child_checkpointing,
        child_checkpoint_self,
        child_checkpoint_done,
        cr_mpirun_abort,
        child_restarting,
    };

    child_context.orterun_argv = malloc((1+argc)*sizeof(char *));
    if (child_context.orterun_argv == NULL) {
        CLOG_PERROR("malloc");
        exit(-1);
    }
    child_context.restart_argv = malloc((1+argc)*sizeof(char *));
    if (child_context.restart_argv == NULL) {
        CLOG_PERROR("malloc");
        exit(-1);
    }

    rc = parse_args(argc, argv, &child_context.orterun_argc, 
        child_context.orterun_argv, &child_context.restart_argc, 
        child_context.restart_argv);
    if (rc < 0) {
        usage();
        exit(rc);
    } else if (rc > 0) {
        /* user asked for help, not an error */
        usage();
        exit(0);
    }

    clog_set_program_name("cr_mpirun");

    CLOG_VERBOSE("Starting main process, pid=%d", getpid());

    /* initialize the event (message) channel */
    channel_init(&event_channel);

    /* set up BLCR */
    rc = (int) cr_init();
    if (rc < 0) {
        exit(-1);
    }

    /* register a handler with BLCR to catch checkpoint signals */
    (void)cr_register_callback(my_callback, &rc, CR_THREAD_CONTEXT);

    /* set up our signal handlers for signal forwarding */
    init_signals(&my_sigmask);

    /* prevent the handler from calling cr_checkpoint() until we let it */
    pthread_mutex_lock(&checkpoint_handler_lock);

    /* create a thread to (fork and) wait for the child */
    child_context.envp = envp;
    child_context.pipe = child_event_pipe;

    rc = pipe(child_context.pipe);
    if (rc < 0) {
        CLOG_PERROR("pipe");
        exit(-1);
    }

    /* initialize child state variables */
    state = child_state_new;

    /* loop until the child exits or something bad happens */
    do {
        CLOG_VERBOSE("state %d: %s", state, state_names[state]);
        state = handle[state](&child_context);
    } while (state != child_state_complete && state != child_state_abort);

    /* abort means that there has been a serious internal error */
    if (state == child_state_abort) {
        CLOG_FATAL("Aborting due to error.");
    } else if (state == child_state_complete) {
        pthread_join(child_context.thread, (void *) &old_status);
        CLOG_DEBUG("Exiting with status %d.", *old_status);
        mimic_exit(*old_status);
        /* shouldn't be reached in normal execution. */
    }

    /* reached on failure cases */
    CLOG_DEBUG("Exiting abnormally.");
    exit(-1);
}
